/* Copyright 2022 Zinc Labs Inc. and Contributors
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
 */

package core

import (
	"context"
	"fmt"
	"os"

	"github.com/aws/aws-sdk-go-v2/config"
	"github.com/aws/aws-sdk-go-v2/service/s3"
	"github.com/blugelabs/bluge"
	"github.com/blugelabs/bluge/analysis"
	"github.com/goccy/go-json"
	"github.com/minio/minio-go/v7"
	"github.com/minio/minio-go/v7/pkg/credentials"
	"github.com/rs/zerolog/log"

	zincanalysis "github.com/zinclabs/zinc/pkg/uquery/v2/analysis"
	"github.com/zinclabs/zinc/pkg/zutils"
)

var systemIndexList = []string{"_index_mapping", "_index_template", "_index", "_metadata", "_users"}

func LoadZincSystemIndexes() (map[string]*Index, error) {
	indexList := make(map[string]*Index)
	for _, index := range systemIndexList {
		log.Info().Msgf("Loading system index... [%s:%s]", index, "disk")
		writer, err := LoadIndexWriter(index, "disk", nil)
		if err != nil {
			return nil, err
		}
		indexList[index] = &Index{
			Name:        index,
			IndexType:   "system",
			StorageType: "disk",
			Writer:      writer,
		}
	}

	return indexList, nil
}

func LoadZincIndexesFromMeta() (map[string]*Index, error) {
	query := bluge.NewMatchAllQuery()
	searchRequest := bluge.NewAllMatches(query).WithStandardAggregations()
	reader, _ := ZINC_SYSTEM_INDEX_LIST["_index"].Writer.Reader()
	defer reader.Close()

	dmi, err := reader.Search(context.Background(), searchRequest)
	if err != nil {
		return nil, fmt.Errorf("core.LoadZincIndexesFromMeta: error executing search: %s", err.Error())
	}

	indexList := make(map[string]*Index)
	next, err := dmi.Next()
	for err == nil && next != nil {
		index := &Index{IndexType: "user", StorageType: "disk"}
		err = next.VisitStoredFields(func(field string, value []byte) bool {
			switch field {
			case "name":
				index.Name = string(value)
			case "index_type":
				index.IndexType = string(value)
			case "storage_type":
				index.StorageType = string(value)
			case "settings":
				json.Unmarshal(value, &index.Settings)
			case "mappings":
				json.Unmarshal(value, &index.CachedMappings)
			default:
			}
			return true
		})

		log.Info().Msgf("Loading user   index... [%s:%s]", index.Name, index.StorageType)
		if err != nil {
			log.Printf("core.LoadZincIndexesFromMeta: error accessing stored fields: %s", err.Error())
		}

		// load index analysis
		if index.Settings != nil && index.Settings.Analysis != nil {
			index.CachedAnalyzers, err = zincanalysis.RequestAnalyzer(index.Settings.Analysis)
			if err != nil {
				log.Printf("core.LoadZincIndexesFromMeta: error parse stored analysis: %s", err.Error())
			}
		}

		// load index data
		var defaultSearchAnalyzer *analysis.Analyzer
		if index.CachedAnalyzers != nil {
			defaultSearchAnalyzer = index.CachedAnalyzers["default"]
		}
		index.Writer, err = LoadIndexWriter(index.Name, index.StorageType, defaultSearchAnalyzer)
		if err != nil {
			log.Error().Msgf("Loading user   index... [%s:%s] index writer error: %s", index.Name, index.StorageType, err.Error())
		}

		// load index docs count
		index.DocsCount, _ = index.LoadDocsCount()

		// load index size
		index.ReLoadStorageSize()

		indexList[index.Name] = index

		next, err = dmi.Next()
	}

	return indexList, nil
}

func LoadZincIndexesFromDisk() (map[string]*Index, error) {
	log.Print("Loading indexes... from disk")

	indexList := make(map[string]*Index)
	dataPath := zutils.GetEnv("ZINC_DATA_PATH", "./data")
	files, err := os.ReadDir(dataPath)
	if err != nil {
		log.Fatal().Msg("Error reading data directory: " + err.Error())
	}

	for _, f := range files {
		iName := f.Name()
		iNameIsSystemIndex := false
		for _, systemIndex := range systemIndexList {
			if iName == systemIndex {
				iNameIsSystemIndex = true
			}
		}
		if iNameIsSystemIndex {
			continue
		}

		tempIndex, err := NewIndex(iName, "disk", NotCompatibleNewIndexMeta, nil)
		if err != nil {
			log.Print("Error loading index: ", iName, " : ", err.Error()) // inform and move in to next index
		} else {
			indexList[iName] = tempIndex
			indexList[iName].IndexType = "user"
			log.Print("Index loaded: " + iName)
		}
	}

	return indexList, nil
}

func LoadZincIndexesFromS3() (map[string]*Index, error) {
	log.Print("Loading indexes from s3...")

	dataPath := zutils.GetEnv("ZINC_S3_BUCKET", "")
	if dataPath == "" {
		return nil, nil
	}

	cfg, err := config.LoadDefaultConfig(context.TODO())
	if err != nil {
		log.Print("Error loading AWS config: ", err)
	}
	client := s3.NewFromConfig(cfg)
	IndexList := make(map[string]*Index)
	delimiter := "/"
	ctx := context.Background()
	params := s3.ListObjectsV2Input{
		Bucket:    &dataPath,
		Delimiter: &delimiter,
	}

	val, err := client.ListObjectsV2(ctx, &params)
	if err != nil {
		log.Print("failed to list indexes in s3: ", err.Error())
		return nil, err
	}

	for _, obj := range val.CommonPrefixes {
		iName := (*obj.Prefix)[0 : len(*obj.Prefix)-1]
		tempIndex, err := NewIndex(iName, "s3", NotCompatibleNewIndexMeta, nil)
		if err != nil {
			log.Print("failed to load index "+iName+" in s3: ", err.Error())
		} else {
			IndexList[iName] = tempIndex
			IndexList[iName].IndexType = "user"
			IndexList[iName].StorageType = "s3"
			log.Print("Index loaded: " + iName)
		}

	}

	return IndexList, nil
}

func LoadZincIndexesFromMinIO() (map[string]*Index, error) {
	log.Print("Loading indexes from minio...")

	endpoint := zutils.GetEnv("ZINC_MINIO_ENDPOINT", "")
	accessKeyID := zutils.GetEnv("ZINC_MINIO_ACCESS_KEY_ID", "")
	secretAccessKey := zutils.GetEnv("ZINC_MINIO_SECRET_ACCESS_KEY", "")
	dataPath := zutils.GetEnv("ZINC_MINIO_BUCKET", "")
	if dataPath == "" {
		return nil, nil
	}

	opts := &minio.Options{
		Creds:  credentials.NewStaticV4(accessKeyID, secretAccessKey, ""),
		Secure: false,
	}

	// Initialize minio client object.
	minioClient, err := minio.New(endpoint, opts)
	if err != nil {
		log.Print(err)
	}

	IndexList := make(map[string]*Index)
	optsList := minio.ListObjectsOptions{
		Recursive: false,
	}

	val := minioClient.ListObjects(context.TODO(), dataPath, optsList)
	if err != nil {
		log.Print("failed to list indexes in minio: ", err.Error())
		return nil, err
	}

	for iName := range val {
		indexName := iName.Key[:len(iName.Key)-1]
		tempIndex, err := NewIndex(indexName, "minio", NotCompatibleNewIndexMeta, nil)
		if err != nil {
			log.Print("failed to load index "+iName.Key+" in minio: ", err.Error())
		} else {
			IndexList[indexName] = tempIndex
			IndexList[indexName].IndexType = "user"
			IndexList[indexName].StorageType = "minio"
			log.Print("Index loaded: " + indexName)
		}
	}

	return IndexList, nil
}

func CloseIndexes() {
	for _, index := range ZINC_INDEX_LIST {
		_ = index.Close()
	}
	for _, index := range ZINC_SYSTEM_INDEX_LIST {
		_ = index.Close()
	}
}
